/*	Copyright (C) 2003-2013 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __houdini_SurfaceAccessibleHoudini_h__
#define __houdini_SurfaceAccessibleHoudini_h__

namespace fe
{

/**************************************************************************//**
    @brief Houdini Surface Binding

	@ingroup houdini
*//***************************************************************************/
class FE_DL_EXPORT SurfaceAccessibleHoudini:
	public SurfaceAccessibleBase
{
	public:
								SurfaceAccessibleHoudini(void):
									m_pGdp(NULL)							{}

virtual							~SurfaceAccessibleHoudini(void)				{}

								//* as SurfaceAccessibleI
virtual	void					bind(Instance a_instance)
								{	setGdp(a_instance.cast<GU_Detail*>()); }
virtual	BWORD					isBound(void)	{ return m_pGdp!=NULL; }

const	std::vector<String>&	groupNames(SurfaceAccessibleI::Element
										a_element) const;

virtual
const	std::vector<SurfaceAccessibleI::Spec>&
								attributeSpecs(SurfaceAccessibleI::Element
										a_element) const;

//virtual	void					copy(sp<SurfaceAccessibleI>
//										a_spSurfaceAccessible);

virtual sp<SurfaceAccessorI>	accessor(SurfaceAccessibleI::Element a_element,
										String a_name,
										SurfaceAccessibleI::Creation a_create);
virtual sp<SurfaceAccessorI>	accessor(SurfaceAccessibleI::Element a_element,
										SurfaceAccessibleI::Attribute
										a_attribute,
										SurfaceAccessibleI::Creation a_create);

virtual	BWORD					discard(SurfaceAccessibleI::Element a_element,
										String a_name);

#ifdef FE_HOUDINI_USE_GA
virtual	sp<SpannedRange>		atomize(AtomicChange a_atomicChange,
										String a_group,U32 a_desiredCount);

virtual void					preparePaging(AtomicChange a_atomicChange,
										String a_group);
virtual I32						pointPage(U32 a_pointIndex) const;
virtual I32						primitivePage(U32 a_primitiveIndex) const;
#endif

virtual	sp<SurfaceI>			surface(String a_group);
virtual	sp<SurfaceI>			subSurface(U32 a_subIndex,String a_group);

virtual	void					append(sp<SurfaceAccessibleI>
										a_spSurfaceAccessibleI,
										const SpatialTransform& a_rTransform);

virtual	void					instance(sp<SurfaceAccessibleI>
										a_spSurfaceAccessibleI,
										const std::vector<SpatialTransform>&
										a_rTransformArray);

								//* Houdini specific
		void					setGdp(const GU_Detail* a_pGdp);
		void					setGdpChangeable(GU_Detail* a_pGdp);
const	GU_Detail*				gdp(void)			{ return m_pGdp; }
		GU_Detail*				gdpChangeable(void)	{ return m_pGdpChangeable; }

	protected:

virtual	void					reset(void);

#ifdef FE_HOUDINI_USE_GA
static	GA_AttributeOwner		attributeOwner(
										SurfaceAccessibleI::Element a_element);
#else
#endif

	private:
//		sp<SpannedRange>		atomizeCrossover(AtomicChange a_atomicChange,
//										String a_group);

		void					harden(sp<SurfaceAccessorHoudini>& rspAccessor);

		sp<SurfaceAccessorHoudini> accessor(void);
		sp<SurfaceAccessorI>	accessorOf(sp<SurfaceAccessorI> a_spAccessor);

		void					initPrimitiveMap(void);
		I32						lookupPrimitive(I32 a_surfaceIndex) const;

const	GU_Detail*							m_pGdp;
		GU_Detail*							m_pGdpChangeable;

		sp<SurfaceI>						m_spSurfaceI;
		std::map<I32,I32>					m_primitiveMap;
		BWORD								m_hasDiscretePolygons;
		std::vector<I32>					m_pageOfPoint;
		std::vector<I32>					m_pageOfPrimitive;

#if FE_HOUDINI_HARDENING
	public:
		sp<SurfaceAccessorHoudini::Hardening>			hardening(String a_key);
	private:
		std::map<String, hp<SurfaceAccessorHoudini::Hardening> > m_hardeningMap;
#endif
};

} // namespace

#endif /* __houdini_SurfaceAccessibleHoudini_h__ */
